import numpy as np
from boris import is_sklearn_available
if is_sklearn_available():
    import sklearn.decomposition as decomposition


class PCA(object):
    """Handmade PCA to prevent sklearn dependency

    """

    def __init__(self, n_components=None):
        self.n_components = n_components
        self.mean = None
        self.w = None
        self.eps = 1e-10

    def fit(self, X):
        """Fit PCA to data in X

        Args:
            X: (np.array) N x D array of datapoints

        Returns:
            PCA object to transform datapoints

        """
        X = X.astype(np.float32)
        self.mean = X.mean(axis=0)
        X = X - self.mean + self.eps
        cov = np.cov(X.T) / X.shape[0]
        v, w = np.linalg.eig(cov)
        idx = v.argsort()[::-1]
        v, w = v[idx], w[:, idx]
        self.w = w
        return self

    def transform(self, X):
        """Use PCA to transform data in X

        Args:
            X: (np.array) N x D array of datapoints

        Returns:
            Array of N x P datapoints where P <= D

        """
        X = X.astype(np.float32)
        X = X - self.mean + self.eps
        return X.dot(self.w)[:, :self.n_components]


def fit_pca(embeddings, n_components=2, fraction=None):
    """Fit PCA to randomly selected subset of embeddings

    Args:
        embeddings: (np.array) 2D array of embeddings with shape N x D
        n_components: (int) number of components to keep
        fraction: (float) fraction (in (0, 1]) of the dataset

    Returns:
        A transformer which can be used to transform embeddings
        to lower dimensions
    """
    N = embeddings.shape[0]
    n = N if fraction is None else min(N, int(N * fraction))
    X = embeddings[np.random.permutation(N)][:n]
    if is_sklearn_available():
        return decomposition.PCA(n_components=n_components).fit(X)
    else:
        return PCA(n_components=n_components).fit(X)



# -*- coding: utf-8 -*-
"""
    boris-cli
    ~~~~~~~~~

    A simple command line tool to train a state-of-the-art
    self-supervised model on your image dataset.

"""

import boris
__requires__ = boris._cli_requires()
if not boris.is_lightning_available():
    msg = 'Using `boris-train` requires the following '
    msg += 'pre-requisites to be installed first: '
    msg += f'{", ".join(__requires__)}.'
    raise RuntimeError(msg)

import hydra
import warnings

from boris.data import ImageCollateFunction
from boris.data import BorisDataset
from boris.embedding import SelfSupervisedEmbedding
from boris.loss import NTXentLoss
from boris.models import ResNetSimCLR

from boris.cli._helpers import _is_url
from boris.cli._helpers import _model_zoo
from boris.cli._helpers import fix_input_path
from boris.cli._helpers import filter_state_dict
from boris.cli._helpers import load_state_dict_from_url

import torch


def _train_cli(cfg, is_cli_call=True):
    """Train a self-supervised model on the image dataset of your choice.

    Args:
        cfg[data]: (str)
            Name of the dataset (to download use cifar10 or cifar100)
        cfg[root]: (str) Directory where the dataset should be stored
        cfg[download]: (bool) Whether to download the dataset
        cfg[from_folder]: (str)
            If specified, the dataset is loaded from the folder

    Returns:
        checkpoint: (str) Path to checkpoint of the best model during training

    """

    data = cfg['data']
    download = cfg['download']

    root = cfg['root']
    if root and is_cli_call:
        root = fix_input_path(root)

    from_folder = cfg['from_folder']
    if from_folder and is_cli_call:
        from_folder = fix_input_path(from_folder)

    if 'seed' in cfg.keys():
        seed = cfg['seed']
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    if torch.cuda.is_available():
        device = 'cuda'
    elif cfg['trainer'] and cfg['trainer']['gpus']:
        device = 'cpu'
        cfg['trainer']['gpus'] = 0

    if cfg['loader']['batch_size'] < 64:
        msg = 'Training a self-supervised model with a small batch size: {}! '
        msg = msg.format(cfg['loader']['batch_size'])
        msg += 'Small batch size may harm embedding quality. '
        msg += 'You can specify the batch size via the loader key-word: '
        msg += 'loader.batch_size=BSZ'
        warnings.warn(msg)

    model = ResNetSimCLR(**cfg['model'])
    checkpoint = cfg['checkpoint']
    if cfg['pre_trained'] and not checkpoint:
        # if checkpoint wasn't specified explicitly and pre_trained is True
        # try to load the checkpoint from the model zoo
        checkpoint, key = _model_zoo(cfg['model'])
        if not checkpoint:
            msg = 'Cannot download checkpoint for key {} '.format(key)
            msg += 'because it does not exist! '
            msg += 'Model will be trained from scratch.'
            warnings.warn(msg)
    elif checkpoint:
        checkpoint = fix_input_path(checkpoint) if is_cli_call else checkpoint
    
    if checkpoint:
        # load the PyTorch state dictionary and map it to the current device
        # then, remove the model. prefix which is caused by the pytorch-lightning
        # checkpoint saver and load the model from the "filtered" state dict
        # this approach is compatible with pytorch_lightning 0.7.1 - 0.8.4 (latest)
        if _is_url(checkpoint):
            state_dict = load_state_dict_from_url(
                checkpoint, map_location=device
            )['state_dict']
        else:
            state_dict = torch.load(
                checkpoint, map_location=device
            )['state_dict']
        if state_dict is not None:
            state_dict = filter_state_dict(state_dict)
            model.load_state_dict(state_dict)

    criterion = NTXentLoss(**cfg['criterion'])
    optimizer = torch.optim.SGD(model.parameters(), **cfg['optimizer'])

    dataset = BorisDataset(root,
                           name=data, train=True, download=download,
                           from_folder=from_folder)

    collate_fn = ImageCollateFunction(**cfg['collate'])
    dataloader = torch.utils.data.DataLoader(dataset,
                                             **cfg['loader'],
                                             collate_fn=collate_fn)

    encoder = SelfSupervisedEmbedding(model, criterion, optimizer, dataloader)
    encoder = encoder.train_embedding(**cfg['trainer'])

    print('Best model is stored at: %s' % (encoder.checkpoint))
    return encoder.checkpoint


@hydra.main(config_path="config/config.yaml", strict=False)
def train_cli(cfg):
    """Train a self-supervised model on the image dataset of your choice.

    Args:
        cfg[data]: (str)
            Name of the dataset (to download use cifar10 or cifar100)
        cfg[root]: (str) Directory where the dataset should be stored
        cfg[download]: (bool) Whether to download the dataset
        cfg[from_folder]: (str)
            If specified, the dataset is loaded from the folder

    Returns:
        checkpoint: (str) Path to checkpoint of the best model during training

    """

    return _train_cli(cfg)


def entry():
    train_cli()



import io
import os

import numpy as np
import warnings
from PIL import Image, ImageFilter

# the following two lines are needed because
# PIL misidentifies certain jpeg images as MPOs
from PIL import JpegImagePlugin
JpegImagePlugin._getmp = lambda: None

LEGAL_IMAGE_FORMATS = ['jpg', 'jpeg', 'png', 'tiff', 'bmp']
MAXIMUM_FILENAME_LENGTH = 80
MAX_WIDTH, MAX_HEIGHT = 2048, 2048


def getenv(key, default):
    """Return the value of the environment variable key if it exists,
       or default if it doesn’t.

    """
    try:
        return os.getenvb(key.encode(), default.encode()).decode()
    except Exception:
        pass
    try:
        return os.getenv(key, default)
    except Exception:
        pass
    return default


def PIL_to_bytes(img, ext='png', quality=None):
    """Return the PIL image as byte stream. Useful to send image via requests.

    """
    bytes_io = io.BytesIO()
    if quality is not None:
        img.save(bytes_io, format=ext, quality=quality)
    else:
        subsampling = -1 if ext.lower() in ['jpg', 'jpeg'] else 0
        img.save(bytes_io, format=ext, quality=100, subsampling=subsampling)
    bytes_io.seek(0)
    return bytes_io


def image_mean(np_img):
    """Return mean of each channel

    """
    return np_img.mean(axis=(0, 1))


def image_std(np_img):
    """Return standard deviation of each channel

    """
    return np_img.std(axis=(0, 1))


def sum_of_values(np_img):
    """Return the sum of the pixel values of each channel

    """
    return np_img.sum(axis=(0, 1))


def sum_of_squares(np_img):
    """Return the sum of the squared pixel values of each channel

    """
    return (np_img ** 2).sum(axis=(0, 1))


def signal_to_noise_ratio(img, axis=None, ddof=0):
    """Calculate the signal to noise ratio of the image

    """
    np_img = np.asanyarray(img)
    mean = np_img.mean(axis=axis)
    std = np_img.std(axis=axis, ddof=ddof)
    return np.where(std == 0, 0, mean / std)


def sharpness(img):
    """Calculate the sharpness of the image using a Laplacian Kernel

    """
    img_bw = img.convert('L')
    filtered = img_bw.filter(
        ImageFilter.Kernel(
            (3, 3),
            # Laplacian Kernel:
            (-1, -1, -1, -1, 8, -1, -1, -1, -1),
            1,
            0,
        )
    )
    return np.std(filtered)


def size_in_bytes(img):
    """Return the size of the image in bytes

    """
    img_file = io.BytesIO()
    img.save(img_file, format=img.format)
    return img_file.tell()


def shape(np_img):
    """Shape of the image as np.ndarray

    """
    return np_img.shape


def get_meta_from_img(img):
    """Calculate the following metadata from a PIL image:
       - Mean
       - Standard Deviation
       - Signal To Noise
       - Sharpness
       - Size in Bytes
       - Shape
       - Sum of Values
       - Sum of Squares

    Args:
        img: (PIL Image)

    Returns:
        A dictionary containing the metadata of the image.
    """

    np_img = np.array(img) / 255.0
    metadata = {
        'mean': image_mean(np_img).tolist(),
        'std': image_std(np_img).tolist(),
        'snr': float(signal_to_noise_ratio(img)),
        'sharpness': float(sharpness(img)),
        'sizeInBytes': size_in_bytes(img),
        'shape': list(shape(np_img)),
        'sumOfValues': sum_of_values(np_img).tolist(),
        'sumOfSquares': sum_of_squares(np_img).tolist(),
    }
    return metadata


def get_thumbnail_from_img(img, size=80):
    """Compute the thumbnail of the image

    Args:
        img: (PIL Image)

    Returns:
        Thumbnail as PIL Image

    """
    thumbnail = img.copy()
    thumbnail.thumbnail((size, size), Image.LINEAR)
    return thumbnail


def resize_image(image, max_width, max_height):
    """

    """
    width = image.width
    height = image.height
    new_image = image.copy()
    new_image.format = image.format

    width_factor = max_width / width
    height_factor = max_height / height
    factor = min(width_factor, height_factor)
    new_image.resize(
        (
            np.floor(factor * width).astype(int),
            np.floor(factor * height).astype(int)
        )
    )
    return new_image


def check_image(filename, mode='full'):
    """Checks whether an image is corrupted or not,
       calculates metadata, and opens file handles for
       image and thumbnail if requested

    Args:
        filename: (str) Path to the file
        mode: (str) In ['full', 'thumbnails', 'none']
            specifies what to upload ('none' is metadata only) 

    Returns:
        A dictionary of metadata and a flag whether file is corrupted
    """

    basename = os.path.basename(filename)
    if len(basename) > MAXIMUM_FILENAME_LENGTH:
        msg = f'Filename {basename} is longer than the allowed maximum of'
        msg += f'{MAXIMUM_FILENAME_LENGTH} character and will be skipped.'
        warnings.warn(msg)
        return None, None, {}, True

    is_corrupted = False
    corruption = ''
    image = Image.open(filename)
    try:
        image.load()
    except IOError as e:
        is_corrupted = True
        corruption = e
    if not is_corrupted and image.format.lower() not in LEGAL_IMAGE_FORMATS:
        is_corrupted = True
        corruption = f'Illegal image format {image.format}.'
    image.close()

    if is_corrupted:
        metadata = {
            'corruption': corruption
        }
    else:
        image = Image.open(filename, 'r')
        metadata = get_meta_from_img(image)
        metadata['corruption'] = ''
        image.close()
    metadata['is_corrupted'] = is_corrupted

    return metadata, is_corrupted


""" Deep Learning Package for Python

boris is a Python module for self-supervised active learning.

"""

try:
    import pytorch_lightning
except ImportError:
    _lightning_available = False
else:
    _lightning_available = True

try:
    import cv2
except ImportError:
    _opencv_available = False
else:
    _opencv_available = True

try:
    import prefetch_generator
except ImportError:
    _prefetch_generator_available = False
else:
    _prefetch_generator_available = True

try:
    import sklearn
except ImportError:
    _sklearn_available = False
else:
    _sklearn_available = True

def _cli_requires():
    return [
        'pytorch_lightning<=0.8.5',
        'opencv-python'
    ]


def is_lightning_available():
    return _lightning_available


def is_opencv_available():
    return _opencv_available


def is_prefetch_generator_available():
    return _prefetch_generator_available


def is_sklearn_available():
    return _sklearn_available


if is_lightning_available():
    from ._one_liners import train_model_and_get_image_features
    from ._one_liners import train_self_supervised_model
    from ._one_liners import get_image_features


def version_compare(v0, v1):
    v0 = [int(n) for n in v0.split('.')][::-1]
    v1 = [int(n) for n in v1.split('.')][::-1]
    pairs = list(zip(v0, v1))[::-1]
    for x, y in pairs:
        if x < y:
            return -1
        if x > y:
            return 1
    return 0


def pretty_print_latest_version(latest_version, width=70):
    lines = [
        'There is a newer version of the package available.',
        'For compatability reasons, please upgrade your current version.',
        '> pip install borisml=={}'.format(latest_version),
    ]
    print('-' * width)
    for line in lines:
        print('| ' + line + (width - len(line) - 3) * " " + "|")
    print('-' * width)


from boris.api import get_latest_version
__version__ = '0.1.10'
latest_version = get_latest_version(__version__)
if latest_version is not None:
    if version_compare(__version__, latest_version) < 0:
        # local version is behind latest version
        pretty_print_latest_version(latest_version)



""" boris.utils

    The boris.utils module provides commonly used utility methods
"""

from ._format_conversion import embeddings_to_pandas
from ._format_conversion import pandas_to_dict
from ._2d_embeddings import fit_pca



import os
import torch

from hydra import utils


def fix_input_path(path):
    """Fix broken relative paths.

    """
    if not os.path.isabs(path):
        path = utils.to_absolute_path(path)
    return path


def filter_state_dict(state_dict):
    """Prevent unexpected key error when loading PyTorch-Lightning checkpoints
       by removing the unnecessary prefix model. from each key.

    """
    new_state_dict = {}
    for key, item in state_dict.items():
        new_key = '.'.join(key.split('.')[1:])
        new_state_dict[new_key] = item
    return new_state_dict


def _is_url(checkpoint):
    is_url = ('https://storage.googleapis.com' in checkpoint)
    return is_url


def _model_zoo(model):

    zoo = {

        'resnet-9/simclr/d16/w0.0625':
        'https://storage.googleapis.com/models_boris/whattolabel-resnet9-simclr-d16-w0.0625-i-ce0d6bd9.pth',

        'resnet-9/simclr/d16/w0.125':
        'https://storage.googleapis.com/models_boris/whattolabel-resnet9-simclr-d16-w0.125-i-7269c38d.pth',

        'resnet-18/simclr/d16/w1.0':
        'https://storage.googleapis.com/models_boris/whattolabel-resnet18-simclr-d16-w1.0-i-58852cb9.pth',

        'resnet-18/simclr/d32/w1.0':
        'https://storage.googleapis.com/models_boris/whattolabel-resnet18-simclr-d32-w1.0-i-085d0693.pth',

        'resnet-34/simclr/d16/w1.0':
        'https://storage.googleapis.com/models_boris/whattolabel-resnet34-simclr-d16-w1.0-i-6e80d963.pth',

    }

    key = model['name']
    key += '/simclr'
    key += '/d' + str(model['num_ftrs'])
    key += '/w' + str(float(model['width']))

    if key in zoo.keys():
        return zoo[key], key
    else:
        return '', key


def load_state_dict_from_url(url, map_location=None):

    try:
        state_dict = torch.hub.load_state_dict_from_url(
            url, map_location=map_location
        )
        return state_dict
    except Exception:
        print('Not able to load state dict from %s' % (url))
        print('Retrying with http:// prefix')
    try:
        url = url.replace('https', 'http')
        state_dict = torch.hub.load_state_dict_from_url(
            url, map_location=map_location
        )
        return state_dict
    except Exception:
        print('Not able to load state dict from %s' % (url))

    # in this case downloading the pre-trained model was not possible
    # notify the user and return
    return {'state_dict': None}
